
#define F_CPU	3686400L

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <string.h>


#include "calibration.h"
#include "colordefinition.h"
#include "com74hc595.h"
#include "uart.h"
  
#define PORTDDR 	DDRB
#define PORTLED		PORTB
#define LED_RED 	PB0
#define LED_GREEN 	PB1
#define LED_BLUE 	PB2


//#define BAUD 2400
#define BAUD 9600
//#define BAUD 38400
//#define BAUD 57600
//#define BAUD 115200

#define uchar unsigned char
#define uint unsigned int
#define	xRC5_IN		PIND
#define	xRC5		PD6			// IR input low active
#define XTAL		3.686400e6
#define RC5TIME 	1.778e-3		// 1.778msec
#define PULSE_MIN	(uchar)(XTAL / 512 * RC5TIME * 0.4 + 0.5)
#define PULSE_1_2	(uchar)(XTAL / 512 * RC5TIME * 0.8 + 0.5)
#define PULSE_MAX	(uchar)(XTAL / 512 * RC5TIME * 1.2 + 0.5)

#define PROG_NO		0
#define PROG_H 		1
#define PROG_S 		2
#define PROG_V 		3
#define PROG_SPEED	4

uchar	rc5_bit;				// bit value
uchar	rc5_time;				// count bit time
uint	rc5_tmp;				// shift bits in
uint	rc5_data;				// store result
uint	rc5_data;				// store result

///////////////////////////////////////////Fernbedienung

uint8_t	eepos;
uint8_t	progmode;

volatile uint8_t do_poll;
volatile uint8_t color_update;
volatile uint8_t color_divide;
volatile unsigned char com74hc595_[ com74hc595_SIZE ];
volatile uint8_t tick;

volatile colorproc colors[1] =
{
	{ 
		{ 
			{ 0xFF, 0xFF, 0xFF, 0x0A, 0, 0, 0, 0, 1, 2 },
		    { 0xFF, 0xFF, 0xFF, 0x0A, 0, 0, 0, 3, 4, 5 } 
		}
	}
};

////////////////////////////////////////////////// BEGIN IRQ

/*
	RC5 Routine
*/
ISR(TIMER0_OVF_vect)
{
  	uint tmp = rc5_tmp;						// for faster access

  	TCNT0 = -2;								// 2 * 256 = 512 cycle

  	if( ++rc5_time > PULSE_MAX )
	{				// count pulse time
    	if( !(tmp & 0x4000) && tmp & 0x2000 )	// only if 14 bits received
      		rc5_data = tmp;
    	tmp = 0;
  	}

  	if( (rc5_bit ^ xRC5_IN) & 1<<xRC5 )
	{		// change detect
    	rc5_bit = ~rc5_bit;						// 0x00 -> 0xFF -> 0x00

    	if( rc5_time < PULSE_MIN )				// to short
      		tmp = 0;

    	if( !tmp || rc5_time > PULSE_1_2 )
		{		// start or long pulse time
      		if( !(tmp & 0x4000) )					// not to many bits
        		tmp <<= 1;							// shift
      		if( !(rc5_bit & 1<<xRC5) )			// inverted bit
        		tmp |= 1;							// insert new bit
      		rc5_time = 0;							// count next pulse time
    	}
  	}

  	rc5_tmp = tmp;
}


/*
	PWM
*/
ISR(TIMER1_OVF_vect) 
{
	TCNT0=0x7F - (tick>>1);

	if (tick < colors[0].colorstrucs[0].red_pwm)
	{
		PORTB |= _BV(LED_RED);
		com74hc595_setBit( colors[0].colorstrucs[0].Pin_R );
		com74hc595_setBit( colors[0].colorstrucs[1].Pin_R );
	}
	else
	{
		PORTB &=~_BV(LED_RED);
		com74hc595_unsetBit( colors[0].colorstrucs[0].Pin_R );
		com74hc595_unsetBit( colors[0].colorstrucs[1].Pin_R );
	}

	if (tick < colors[0].colorstrucs[0].green_pwm)
	{
		PORTB |= _BV(LED_GREEN);
		com74hc595_setBit( colors[0].colorstrucs[0].Pin_G );
		com74hc595_setBit( colors[0].colorstrucs[1].Pin_G );
	}
	else
	{
		PORTB &=~_BV(LED_GREEN);
		com74hc595_unsetBit( colors[0].colorstrucs[0].Pin_G );
		com74hc595_unsetBit( colors[0].colorstrucs[1].Pin_G );
	}

	if (tick < colors[0].colorstrucs[0].blue_pwm)
	{
		PORTB |= _BV(LED_BLUE);
		com74hc595_setBit( colors[0].colorstrucs[0].Pin_B );
		com74hc595_setBit( colors[0].colorstrucs[1].Pin_B );
	}
	else
	{
		PORTB &=~_BV(LED_BLUE);
		com74hc595_unsetBit( colors[0].colorstrucs[0].Pin_B );
		com74hc595_unsetBit( colors[0].colorstrucs[1].Pin_B );
	}

	tick++;
}

/*
	PWM
*/
ISR(TIMER2_OVF_vect) 
{
	TCNT2 = -128;
	do_poll = 1;
	color_divide++;
	if (color_divide==0) 
	{
		color_divide=0x7F;
		color_update++;
	}
}

////////////////////////////////////////////////// END IRQ

void hsv_to_rgb(void) 
{
	uint8_t r,g,b, i, f;
	uint8_t p, q, t;

	if( colors[0].colorstrucs[0].S == 0 ) 
	{	
		r = g = b = colors[0].colorstrucs[0].V;
	}
	else
	{	
		i = colors[0].colorstrucs[0].H / 43;
		f = colors[0].colorstrucs[0].H % 43;
		p = ( colors[0].colorstrucs[0].V * ( 255 - colors[0].colorstrucs[0].S ) ) / 256;
		q = ( colors[0].colorstrucs[0].V * ( ( 10710 - ( colors[0].colorstrucs[0].S * f ) ) / 42 ) ) / 256;
		t = ( colors[0].colorstrucs[0].V * ( ( 10710 - ( colors[0].colorstrucs[0].S * ( 42 - f) ) ) / 42 ) ) / 256;

		switch( i )
		{	
			case 0:		r = colors[0].colorstrucs[0].V; g = t; b = p; break;
			case 1:		r = q; g = colors[0].colorstrucs[0].V; b = p; break;
			case 2:		r = p; g = colors[0].colorstrucs[0].V; b = t; break;
			case 3:		r = p; g = q; b = colors[0].colorstrucs[0].V; break;			
			case 4:		r = t; g = p; b = colors[0].colorstrucs[0].V; break;				
			case 5:
			default:	r = colors[0].colorstrucs[0].V; g = p; b = q; break;
		}
	}

	r = get_calibration( red_calibration, r );
	g = get_calibration( green_calibration, g );
	b = get_calibration( blue_calibration, b );
	cli();                                              // Kein Interrupt beim Übernehmen der neuen Werte!
	colors[0].colorstrucs[0].red_pwm = r; 
	colors[0].colorstrucs[0].green_pwm = g; 
	colors[0].colorstrucs[0].blue_pwm = b;
	sei();
}

/////// Predefined Colors
EEMEM colorproc colortable[8] =
{
		{ 
		  { 
		    // White 
		    { 0x00, 0x00, 0xFF, 200, 0, 0, 0, 1, 2, 3 },
		    { 0x00, 0x00, 0xFF, 200, 0, 0, 0, 4, 5, 6 } 
		  }
        }, 
		{ 
		  { 
		    // Bright red
			{ 0x00, 0xFF, 0xFF, 0xC8, 0, 0, 0, 1, 2, 3 },
		    { 0x00, 0xFF, 0xFF, 0xC8, 0, 0, 0, 4, 5, 6 } 
		  }
        }, 
		{ 
		  { 
		    // Bright yellow
			{ 0x2A, 0xFF, 0xFF, 0xC8, 0, 0, 0, 1, 2, 3 },
		    { 0x2A, 0xFF, 0xFF, 0xC8, 0, 0, 0, 4, 5, 6 } 
		  }
        }, 
		{ 
		  { 
		    // Bright green
			{ 0x55, 0xFF, 0xFF, 0xC8, 0, 0, 0, 1, 2, 3 },
		    { 0x55, 0xFF, 0xFF, 0xC8, 0, 0, 0, 4, 5, 6 } 
		  }
        }, 
		{ 
		  { 
		    // Bright cyan
			{ 0x80, 0xFF, 0xFF, 0xC8, 0, 0, 0, 1, 2, 3 },
		    { 0x80, 0xFF, 0xFF, 0xC8, 0, 0, 0, 4, 5, 6 } 
		  }
        }, 
		{ 
		  { 
		    // Bright blue
			{ 0xAA, 0xFF, 0xFF, 0xC8, 0, 0, 0, 1, 2, 3 },
		    { 0xAA, 0xFF, 0xFF, 0xC8, 0, 0, 0, 4, 5, 6 } 
		  }
        }, 
		{ 
		  { 
		    // fast cycle
			{ 0x00, 0xFF, 0xFF, 0x0A, 0, 0, 0, 1, 2, 3 },
		    { 0x00, 0xFF, 0xFF, 0x0A, 0, 0, 0, 4, 5, 6 } 
		  }
        }, 
		{ 
		  { 
		    // fast cycle low sat
			{ 0x00, 0x7F, 0xFF, 0x32, 0, 0, 0, 3, 4, 5 },
		    { 0x00, 0x7F, 0xFF, 0x32, 0, 0, 0, 4, 5, 6 } 
		  }
        }
};

uint8_t getKeycode( uint8_t device, uint8_t key )
{
	uint8_t result = 0xFF;
	if ( ( device == 0x14 ) && ( key == 0x00 ) )
	{
		return 0x00;
	};
	if ( ( device == 0x14 ) && ( key == 0x01 ) )
	{
		return 0x01;
	};
	if ( ( device == 0x14 ) && ( key == 0x02 ) )
	{
		return 0x02;
	};
	if ( ( device == 0x14 ) && ( key == 0x03 ) )
	{
		return 0x03;
	};
	if ( ( device == 0x14 ) && ( key == 0x04 ) )
	{
		return 0x04;
	};
	if ( ( device == 0x14 ) && ( key == 0x05 ) )
	{
		return 0x05;
	};
	if ( ( device == 0x14 ) && ( key == 0x06 ) )
	{
		return 0x06;
	};
	if ( ( device == 0x14 ) && ( key == 0x07 ) )
	{
		return 0x07;
	};
	if ( ( device == 0x14 ) && ( key == 0x08 ) )
	{
		return 0x08;
	};
	if ( ( device == 0x14 ) && ( key == 0x09 ) )
	{
		return 0x09;
	};

	if ( ( device == 0x11 ) && ( key == 0x10 ) )
	{
		return 0x11;
	};
	if ( ( device == 0x11 ) && ( key == 0x11 ) )
	{
		return 0x12;
	};
	if ( ( device == 0x11 ) && ( key == 0x20 ) )
	{
		return 0x13;
	};
	if ( ( device == 0x11 ) && ( key == 0x21 ) )
	{
		return 0x14;
	};
	return result;
}

////////////////////////

void rc5_handler( uint i ) 
{
	uint8_t device;
	uint8_t key;
	uint8_t keycode;

	static uint8_t speed;
	static uint16_t last_data;



	if (i == last_data) 
	{
		if (speed < 20)
			speed++;
	} 
	else 
	{
		speed = 1;
		last_data = i;
	}

	device = i >> 6 & 0x1F;			// Device-Code auslesen
	key = (i & 0x3F);				// Key-Code auslesen
	keycode = getKeycode( device, key );

uart_write_P(PSTR("keycode:"));
uart_hex(keycode);

	if ( keycode <= 0x09 )
	{
		eeprom_read_block( (void*)&colors[0], (const void*)&colortable[keycode], sizeof( colorproc ) );
	}
/*
	else if ( keycode == 0x00 )				// Speichern
	{
		progmode = PROG_NO;
	}
	else if ( keycode == 0x01 )
	{
		progmode = PROG_H;
	}
	else if ( keycode == 0x02 )
	{
		progmode = PROG_S;
	}
	else if ( keycode == 0x03 )
	{
		progmode = PROG_V;
	}
	else if ( keycode == 0x04 )
	{
		progmode = PROG_SPEED;
	}
*/
	else if ( keycode == 0x13 ) 		// Channel +
	{
		eepos++;    
		if ( eepos == NUMBER_OF_RGB_TABLE )
		{
			eepos = 0;
		}
		eeprom_read_block( (void*)&colors[0], (const void*)&colortable[eepos], sizeof( colorproc ) );
	}
	else if ( keycode == 0x14 ) 	// Channel -
	{
		eepos--;    
		if ( eepos == 0 )
		{
			eepos = NUMBER_OF_RGB_TABLE - 1;
		}
		eeprom_read_block( (void*)&colors[0], (const void*)&colortable[eepos], sizeof( colorproc ) );
	}
	else if ( ( keycode==0x11 ) && ( progmode == PROG_NO ) )	// Helligkeit +
	{
		if ( colors[0].colorstrucs[0].V <= 0xFF - speed) 
		{
			colors[0].colorstrucs[0].V += speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].V =	0xFF;
		}
	} 
	else if ( ( keycode==0x12 ) && ( progmode == PROG_NO ) )	// Helligkeit -
	{
		if ( colors[0].colorstrucs[0].V >= speed ) 
		{
			colors[0].colorstrucs[0].V -= speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].V = 0;
		}
	} 
	else if ( ( keycode==0x11 ) && ( progmode == PROG_H ) )
	{
		if ( colors[0].colorstrucs[0].H <= 0xFF - speed ) 
		{
			colors[0].colorstrucs[0].H += speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].H =	0xFF;
		}
	}
	else if ( ( keycode==0x12 ) && ( progmode == PROG_H ) )
	{
		if ( colors[0].colorstrucs[0].H >= speed ) 
		{
			colors[0].colorstrucs[0].H -= speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].H = 0;
		}
	}
	else if ( ( keycode==0x11 ) && ( progmode == PROG_S ) )
	{
		if ( colors[0].colorstrucs[0].S <= 0xFF - speed ) 
		{
			colors[0].colorstrucs[0].S += speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].S = 0xFF;
		}
	} 
	else if ( ( keycode==0x12 ) && ( progmode == PROG_S ) )
	{
		if ( colors[0].colorstrucs[0].S >= speed ) 
		{
			colors[0].colorstrucs[0].S -= speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].S = 0;
		}
	}
	else if ( ( keycode==0x11 ) && ( progmode == PROG_V ) )
	{
		if ( colors[0].colorstrucs[0].V <= 0xFF - speed ) 
		{
			colors[0].colorstrucs[0].V += speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].V =	0xFF;
		}
	} 
	else if ( ( keycode==0x12 ) && ( progmode == PROG_V ) )
	{
		if ( colors[0].colorstrucs[0].V >= speed ) 
		{
			colors[0].colorstrucs[0].V -= speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].V = 0;
		}
	} 
	else if ( ( keycode==0x11 ) && ( progmode == PROG_SPEED ) )
	{
		if ( colors[0].colorstrucs[0].Speed <= 0xFF - speed ) 
		{
			colors[0].colorstrucs[0].Speed += speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].Speed = 0xFF;
		}
	} 
	else if ( ( keycode==0x12 ) && ( progmode == PROG_SPEED ) )
	{
		if ( colors[0].colorstrucs[0].Speed >= speed ) 
		{
			colors[0].colorstrucs[0].Speed -= speed;
		} 
		else 
		{
			colors[0].colorstrucs[0].Speed = 0;
		}
	} 
	rc5_data=0;
}

void initTimer( void )
{
	TCCR0  = 0;
	TCCR0  = 1<<CS02;			         	

	TCCR1A = 1<<CS10;			         	
	TCCR1B = 1<<CS10;			         	

	TCCR2  = 0;
	TCCR2  = 1<<CS11;			         	

 	TIMSK  = 1<<TOIE0^1<<TOIE1^1<<TOIE2;	// enable timer interrupt
}

int main(void) 
{
 	uint i;									// Variable mit gltigem RC5-Code
	
	initTimer();
//  uart_init(UART_BAUD_CALC(BAUD));
//	com74hc595_init();

	DDRD = 0x00;

	progmode = PROG_NO;
	eepos 	= 2;

//	eeprom_read_block( (void*)&colors[0], (const void*)&colortable[eepos], 20 );

	hsv_to_rgb();  

	sei();

	while ( 1 ) 
	{
    	cli();
    	i = rc5_data;						// lese die zwei Bytes fr die RC5-Codierung
    	rc5_data = 0;
//		com74hc595_out();
		sei();
    	if( i )
    	{
			rc5_handler( i );
			hsv_to_rgb();  
		} 
		else if ( colors[0].colorstrucs[0].Speed != 0xFF && ( color_update > colors[0].colorstrucs[0].Speed ) ) 
		{ // "else": Prevent high update rate from messing with RC5...
			color_update = 0;
			colors[0].colorstrucs[0].H++;
			hsv_to_rgb();  
		}
	} 
	return 0;
}

